/*
 * Copyright (C) 1998-2002 <a href="mailto:jens.dietrich@unforgettable.com">Jens Dietrich</a>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package de.jfs2002.mandarax.crm;

import org.mandarax.kernel.*;
import org.mandarax.reference.ResolutionInferenceEngine;
import org.mandarax.util.LogicFactorySupport;

/**
 * The rule based disount calculator. Internally, the discount is calculated by issuing a query
 * to the inference engine. This class encapsulates the inference engine and most of
 * the mandarax logic, in a EJB context this would be a classic session bean.
 * @author <A HREF="mailto:jens.dietrich@unforgettable.com">Jens Dietrich</A>
 * @version 1.8 <18 June 2002>
 * @since 1.2
 */
public class DiscountCalculator extends LogicFactorySupport {

	private ResultSet rs = null;

	// the name of the variable we use in queries .. could be any string
	private static final String QUERY_VARIABLE = "DISCOUNT?";

	/**
	 * Constructor.
	 */
	public DiscountCalculator() {
		super();
	}

	/**
	 * Calculate the discount for a customer c using the knowledge base kb.
	 * @param c the customer
	 * @param kb the knowledge base
	 */
	public void calculateDiscount(Customer c, KnowledgeBase kb) {

		// build query
		Predicate p = KBServer.getPredicateGetDiscount();
		Fact query = fact(p, c, QUERY_VARIABLE);

		// issue query
		InferenceEngine ie = new ResolutionInferenceEngine();

		try {
			rs = ie.query(query, kb, ie.ONE, ie.TRY_NEXT);
			rs.first ();
		}
		catch (InferenceException x) {
			System.err.println("Inference Error, details:");
			x.printStackTrace(System.err);
		}
	}

	/**
	 * Get the derivation used to calculate the result
	 * @return the derivation used to calculate the discount
	 */
	public Derivation getDerivation() {
	    try {
			return (rs == null) ? null : (Derivation) rs.getProof();
		} catch (InferenceException ex) {
			System.err.println("Inference Error, details:");
			ex.printStackTrace(System.err);
		}
		return null;
	}

	/**
	 * Get the calculated discount
	 * @return the calculated discount
	 */
	public Discount getDiscount() {
		try {
			return (rs== null) ? null : (Discount) rs.getResult(Discount.class, QUERY_VARIABLE);
		} catch (InferenceException ex) {
			System.err.println("Inference Error, details:");
			ex.printStackTrace(System.err);
		}
		return null;
	}

	/**
	 * Reset the calculator. Should be used between queries if the same instance is used
	 * for multiple queries.
	 */
	public void reset() {
		rs = null;
	}
}