/*
 * Copyright (C) 1998-2002 <a href="mailto:jens.dietrich@unforgettable.com">Jens Dietrich</a>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package de.jfs2002.mandarax.crm;


import java.util.Collection;
import java.util.Hashtable;

import org.mandarax.kernel.Fact;
import org.mandarax.kernel.Function;
import org.mandarax.kernel.KnowledgeBase;
import org.mandarax.kernel.Predicate;
import org.mandarax.kernel.Rule;
import org.mandarax.kernel.SimplePredicate;
import org.mandarax.kernel.meta.JFunction;
import org.mandarax.lib.math.DoubleArithmetic;
import org.mandarax.reference.AdvancedKnowledgeBase;
import org.mandarax.util.AutoFacts;
import org.mandarax.util.LogicFactorySupport;

/**
 * Simulation of a knowledge base. I.e., a test knowledge base is hardcoded here.
 * In a real world scenario, this class would probably be replaced by a class
 * reading (and perhaps caching) data from a RDBMS or from a XML source.
 * Our intention is saving logic as data!
 * <br> There is also a simple mechanism provided to serve aliases for the
 * rules.
 * @author <A HREF="mailto:jens.dietrich@unforgettable.com">Jens Dietrich</A>
 * @version 1.8 <18 June 2002>
 * @since 1.2
 */
public class KBServer extends LogicFactorySupport {

    private KnowledgeBase kb            = null;
    private Hashtable     aliasRegistry = null;

    /**
     * Get the alias name for an object (rule or fact).
     * @return an alias name
     * @param obj an object
     */
    public String getAlias(Object obj) {
        String alias = (String) getAliasRegistry ().get (obj);

        if(alias == null) {
            if(obj instanceof Fact) {
                alias = "Fact generated from transaction data";
            } else if(obj instanceof AutoFacts) {
                alias =
                    "Autofacts - facts will be built dynamically from transactions";
            } else {
                alias = "null";
            }
        }

        return alias;
    }

    /**
     * Get the alias registry.
     * @return the registry
     */
    private Hashtable getAliasRegistry() {
        if(aliasRegistry == null) {

            // force initialization
            getKB ();
        }
        ;

        return aliasRegistry;
    }

    /**
     * Get the "get turnover" function
     * @return a function
     */
    private static Function getFunctionTurnover() {
        try {
            Class[] par = { Integer.TYPE };

            return new JFunction (
                Customer.class.getMethod ("getTurnover", par), "turnover");
        } catch(NoSuchMethodException x) {
            x.printStackTrace (System.err);
        }

        return null;
    }

    /**
     * Get the "get turnover made with a certain kind of payment" function
     * @return a function
     */
    private static Function getFunctionTurnoverPayment() {
        try {
            Class[] par = { Integer.TYPE, KindOfPayment.class };

            return new JFunction (
                Customer.class.getMethod ("getTurnover", par),
                "turnover/payment");
        } catch(NoSuchMethodException x) {
            x.printStackTrace (System.err);
        }

        return null;
    }

    /**
     * Get the sample knowledge base. The knowledge base is hardcoded here,
     * but should actually come from a data source.
     * @return a knowledge base
     */
    public KnowledgeBase getKB() {
        if(kb == null) {

            // useful variable .. the name of a customer type variable
            String c = "a customer";

            // do lazy initialization
            kb            = new AdvancedKnowledgeBase ();
            aliasRegistry = new Hashtable ();

            Rule rule1 =
                rule (fact (
                    DoubleArithmetic.GREATER_THAN, cplx (
                        getFunctionTurnoverPayment (), c, new Integer (
                            12), DB.getKindOfPayment (
                            KindOfPayment.COMPANY_VISA)), new Double (
                                200)), fact (
                                    getPredicateHasCategory (), c, DB
                                        .getCategory (Category.PLATINUM)));
            Rule rule2 =
                rule (fact (
                    DoubleArithmetic.GREATER_THAN, cplx (
                        getFunctionTurnover (), c, new Integer (
                            12)), new Double (250)), fact (
                                getPredicateHasCategory (), c, DB
                                    .getCategory (Category.GOLD)));
            Rule rule3 =
                rule (fact (
                    getPredicateHasCategory (), c, DB.getCategory (
                        Category.PLATINUM)), fact (
                            getPredicateGetDiscount (), c, new Discount (
                                12.5)));
            Rule rule4 =
                rule (fact (
                    getPredicateHasCategory (), c, DB.getCategory (
                        Category.GOLD)), fact (
                            getPredicateGetDiscount (), c, new Discount (
                                7.5)));
            Rule rule5 =
                rule (fact (
                    DoubleArithmetic.GREATER_THAN, cplx (
                        getFunctionTurnover (), c, new Integer (
                            12)), new Double (100)), fact (
                                getPredicateGetDiscount (), c, new Discount (
                                    5)));

            // last rule is actually a fact! .. each customer deserves a discount
            Fact rule6 = fact (getPredicateGetDiscount (), c,
                               new Discount (2.5));

            // and finally auto facts. this autofacts are a convenient way to integrate facts
            // built from the database into the knowledge base
            AutoFacts af = new AutoFacts () {

                public Collection getExtension(Class clazz) {
                    if(clazz == Customer.class) {
                        return DB.getCustomers ().values ();
                    }

                    if(clazz == Category.class) {
                        return DB.getCategories ().values ();
                    }

                    if(clazz == KindOfPayment.class) {
                        return DB.getKindsOfPayment ().values ();
                    }

                    return super.getExtension (clazz);
                }
            };
            Predicate[] predicates = { DoubleArithmetic.GREATER_THAN };

            af.setPredicates (predicates);
            kb.add (rule1);
            kb.add (rule2);
            kb.add (rule3);
            kb.add (rule4);
            kb.add (rule5);
            kb.add (rule6);
            kb.add (af);
            aliasRegistry.put (
                rule1,
                "if the turnover of a customer in the last 12 months payed with COMPANY VISA is more than US$ 200.00 then a customer has customer category Platinum");
            aliasRegistry.put (
                rule2,
                "if the turnover of a customer in the last 12 months is more than US$ 250.00 then a customer has customer category Gold");
            aliasRegistry.put (
                rule3,
                "if a customer has customer category Platinum then a customer gets a discount of 12.5%");
            aliasRegistry.put (
                rule4,
                "if a customer has customer category Gold then a customer gets a discount of 7.5%");
            aliasRegistry.put (
                rule5,
                "if the turnover of a customer in the last 12 months is more than US$ 100.00 then a customer gets a discount of 5.0%");
            aliasRegistry.put (rule6,
                               "each customer gets a discount of 2.5%");
        }

        return kb;
    }

    /**
     * Get the "customer gets a certain discount" predicate
     * @return a predicate
     */
    public static Predicate getPredicateGetDiscount() {
        Class[] par = { Customer.class, Discount.class };

        return new SimplePredicate ("getsDiscount", par);
    }

    /**
     * Get the "customer is in a certain category" predicate
     * @return a predicate
     */
    private static Predicate getPredicateHasCategory() {
        Class[] par = { Customer.class, Category.class };

        return new SimplePredicate ("hasCategory", par);

        //Class[] par2 = {Customer.class,Discount.class};
        //predicates[1] = new SimplePredicate("getsDiscount",par2);
    }
}
